MODULE stroumpf
  USE utils
  IMPLICIT NONE
  REAL, PARAMETER:: pi=ACOS(-1.)
  PRIVATE
  PUBLIC:: init,res, vmax
CONTAINS
  !
  ! Evaluation of residuals
  ! R. Abgrall, remi.abgrall@math.uzh.ch

  FUNCTION init(x,gamma,n) RESULT(u)
    IMPLICIT NONE

    INTEGER,INTENT(in):: n
    REAL, INTENT(in):: gamma 
    REAL, INTENT(in):: x(0:n+1)
    REAL, DIMENSION(3,0:n+1):: u
    INTEGER :: i
    REAL :: dx,y
    REAL, DIMENSION(2):: ro, v, p


    ! SOD
    ro(1)=1;ro(2)=0.125
    v(1)=0; v(2)=0
    p(1)=1;p(2)=0.1
!!$    !$!! LAX
!!$     ro(1)=0.445;ro(2)=0.5!.125
!!$       v(1)=0.698; v(2)=0
!!$       p(1)=3.528;p(2)=0.571!0.1

    !
    u(1,:)=ro(1)
    u(2,:)=ro(1)*v(1)
    u(3,:)=p(1)/(gamma-1)+0.5*ro(1)*v(1)*v(1)
    DO i=0,n+1
       IF (x(i)>0.5) THEN
          u(1,i)=ro(2)
          u(2,i)=ro(2)*v(2)
          u(3,i)=p(2)/(gamma-1)+0.5*ro(2)*v(2)*v(2)
       ENDIF
    ENDDO

    ! Blast wave
!!$    DO i=0,n+1
!!$       IF (x(i)<0.1) THEN
!!$          u(1,i)=1.
!!$          u(2,i)=0.
!!$          u(3,i)= 1000./(gamma-1.)
!!$       ELSE
!!$          IF (x(i)<0.9) THEN
!!$             u(1,i)=1.
!!$             u(2,i)=0.
!!$             u(3,i)= 0.1/(gamma-1.)
!!$          ELSE
!!$             u(1,i)=1.
!!$             u(2,i)=0.
!!$             u(3,i)= 100./(gamma-1.)
!!$          ENDIF
!!$       ENDIF
!!$enddo
!!$    DO i=0,n+1
!!$       IF (x(i)<0.5) THEN
!!$          u(1,i)=ro(1)
!!$          u(2,i)=ro(1)*v(1)
!!$          u(3,i)=p(1)/0.4+0.5*ro(1)*v(1)*v(1)
!!$       ELSE
!!$          u(1,i)=ro(2)
!!$          u(2,i)=ro(2)*v(2)
!!$          u(3,i)=p(2)/0.4+0.5*ro(2)*v(2)*v(2)
!!$       ENDIF
!!$    ENDDO


!!$    DO i=0,n+1
!!$       y=(x(i)-0.5)/0.2
!!$       ro(1)=1.
!!$       v(1)=0.
!!$       p(1)=1.
!!$       if (x(i).le.0.5) then
!!$          v(1)=-0.1
!!$       else
!!$          v(1)=0.1
!!$       endif
!!$            IF (x(i).GE.0.2.AND.x(i).LE.0.8) THEN
!!$        v(1)=-0.1*( x(i)-0.8)/(0.2-0.8)+0.1*(x(i)-0.2)/(0.8-0.2)
!!$     ENDIF
!!$ !      if (y.ge.-1.and. y.le.1.) then
!!$ !         v(1)=sin(pi*y)**5
!!$  !     endif
!!$       u(1,i)=ro(1)
!!$       u(2,i)=ro(1)*v(1)
!!$       u(3,i)=p(1)/0.4+0.5*ro(1)*v(1)*v(1)
!!$
!!$    ENDDO

!!$    U1(:)=u(:,1)
!!$    U2(:)=u(:,n+1)
  END FUNCTION init

  ! Evaluation of residual
  SUBROUTINE res(ischema,n,iordre,un,u,nubo,dx,dt,gamma,sp)
    IMPLICIT NONE
    !
    ! This version uses the approximation by the N scheme for space
    ! u --> u^{n+1}
    ! un--> u^n
    !
    INTEGER, INTENT(in) :: n, iordre, ischema
    REAL, INTENT(in) :: dx, dt, gamma
    REAL, DIMENSION(3,0:n+1),INTENT(inout):: un,u! 3*N
    INTEGER, DIMENSION(2,1:n),INTENT(in) :: nubo
    REAL, DIMENSION(3,0:n+1), INTENT(out):: sp
    REAL ,DIMENSION(3):: ub,ubn, phi1, phi2, phin1, phin2!, v1, v2
    REAL,DIMENSION(2,3):: resi
    REAL, DIMENSION(2,3) :: beta
    INTEGER, DIMENSION(2,2):: nub
    INTEGER, DIMENSION(2) :: i
    REAL, DIMENSION(3,3) :: u0
    REAL, DIMENSION(3) :: phii
    REAL, DIMENSION(3,3):: xk1, xk2, xk1n, xk2n
    INTEGER :: i1,i2, iseg, kiseg
    REAL :: xn, rap1


    sp=0
    DO iseg=1,n  ! schema n a l'instant t_{n+1}
       i1=nubo(1,iseg);i2=nubo(2,iseg)

       ! original
       CALL phi(ischema,u (:,i1),u (:,i2),gamma,phi1 ,phi2 )
       CALL phi(ischema,un(:,i1),un(:,i2),gamma,phin1,phin2)
       resi(1,:)=0.5*dt*(phi1+phin1)+dx*(un(:,i1)-u(:,i1))*0.5
       resi(2,:)=0.5*dt*(phi2+phin2)+dx*(un(:,i2)-u(:,i2))*0.5


       ! First or second order

       CALL lim(iordre,gamma,u(:,i1),u(:,i2),resi,beta)

       sp(:,i1)=sp(:,i1)+ (beta(1,:))
       sp(:,i2)=sp(:,i2)+ (beta(2,:))

    ENDDO
    sp(:,0)=0;sp(:,n+1)=0.
  END SUBROUTINE res



  SUBROUTINE lim(iordre,gamma,u1,u2,resi,beta)
    IMPLICIT NONE
    ! "limitation"
    INTEGER, INTENT(in) :: iordre
    REAL ,DIMENSION(2,3), INTENT(in):: resi
    REAL,DIMENSION(2,3), INTENT(out):: beta
    REAL, DIMENSION(3), INTENT(in):: u1, u2
    REAL, INTENT(in):: gamma
    REAL  :: phi 
    REAL ,DIMENSION(3):: r0, r1, r2, vp
    REAL :: ub, hb, ro1, ro2, v1, v2, h1, h2 , den, c2, c
    REAL :: p1, p2, p, c1, pp1, pp2,pp
    REAL :: kappa, x1 , x2, blend
    REAL, DIMENSION(2):: aa, bb
    REAL, DIMENSION(2,3):: limi, l
    INTEGER :: k
    kappa=gamma-1

    ro1=u1(1)
    ro2=u2(1)
    v1=u1(2)/u1(1)
    v2=u2(2)/u2(1)
    p1=kappa*( u1(3)-0.5*ro1*v1*v1)
    p2=kappa*( u2(3)-0.5*ro2*v2*v2)
    h1=(u1(3)+p1)/ro1
    h2=(u2(3)+p2)/ro2
    c1=SQRT(gamma*p1/ro1)
    c2=SQRT(gamma*p2/ro2)
    IF (iordre.EQ.1) THEN
       beta=resi
    ELSE
       !
       ! Using Roe average (other simpler version exist:
       ! this is not really mandatory)
       !


       den=SQRT(ro1)/(SQRT(ro1)+SQRT(ro2))
       ub= den*v1+(1-den)*v2
       hb= den*h1+(1-den)*h2
       c2=kappa*( hb- 0.5*ub*ub )
       c=SQRT(c2)
       vp(1)=ub; vp(2)=ub-c; vp(3)=ub+c

       ! Compute characteristic decomposition or first order residuals
       !
       aa(:)=kappa/c2*( resi(:,3)-ub*resi(:,2)+0.5*ub*ub*resi(:,1))
       bb(:)=(resi(:,2)-ub*resi(:,1))/c
       l(:,1)=resi(:,1)-aa(:)
       l(:,2)=0.5*( aa(:)-bb(:) )
       l(:,3)=0.5*( aa(:)+bb(:) )

       r0(1)=1.; r0(2)=ub  ; r0(3)=ub*ub*0.5
       r1(1)=1.; r1(2)=ub-c; r1(3)=hb-ub*c
       r2(1)=1.; r2(2)=ub+c; r2(3)=hb+ub*c
       !
       ! "limit"
       !
       DO k=1,3
          pp1=l(1,k)
          pp2=l(2,k)
          pp=pp1 + pp2
          ! psi
          IF (ABS(pp)>0.) THEN 
             den=MAX(pp1/pp,0.)+MAX(pp2/pp,0.)
             limi(1,k)=MAX(pp1/pp,0.)/den*pp
             limi(2,k)=MAX(pp2/pp,0.)/den*pp

          ELSE
             limi(:,k)=0.
          ENDIF

       ENDDO


       beta(1,:)=limi(1,1)*r0+limi(1,2)*r1+limi(1,3)*r2
       beta(2,:)=limi(2,1)*r0+limi(2,2)*r1+limi(2,3)*r2
    ENDIF


  END SUBROUTINE lim

  SUBROUTINE phi(ischema,U1,U2, gamma,phi1,phi2)
    IMPLICIT NONE
    !
    ! compute phi1=A^+(U_2-U_1), Phi2=A^-(U2-U1) or LxF residual
    !  A computed for  Ub, Roe average. can also be arithemitic average
    !
    INTEGER, INTENT(in):: ischema
    REAL, DIMENSION(3), INTENT(in):: U1,U2
    REAL, INTENT(in) :: gamma
    REAL, INTENT(out),DIMENSION(3):: phi1, phi2
    REAL, DIMENSION(3):: ph
    REAL, DIMENSION(3):: vp,l
    REAL, DIMENSION(3)::  r0, r1, r2, psi
    REAL:: aa,bb, kappa
    REAL:: v, p, H, c2, c, den
    REAL:: p2,p1, v2,v1, ro2, ro1, alpha

    ! Roe 
    kappa=gamma-1.

    ro1=u1(1)
    v1 =u1(2)/u1(1)
    p1 =kappa*( u1(3)-0.5*ro1*v1*v1)

    ro2=u2(1)
    v2 =u2(2)/u2(1)
    p2 =kappa*(u2(3)-0.5*ro2*v2*v2)
    !
    ! average
    !
    den=1./(SQRT(ro2)+SQRT(ro1))
    v=(SQRT(ro2)*v2+SQRT(ro1)*v1)*den
    H= ((u2(3)+p2)/SQRT(ro2)+(u1(3)+p1)/SQRT(ro1))*den
    c2=kappa*(H-0.5*v*v)
    c=SQRT(c2)

    vp(1)=v; vp(2)=v-c; vp(3)=v+c


    aa= (p2-p1)/c2
    bb=SQRT(ro1*ro2)*(v2-v1)/c

    l(1)=(ro2-ro1)-aa
    l(2)=0.5*(  aa-bb )
    l(3)=0.5*(  aa+bb )

    r0(1)=1.; r0(2)=v  ; r0(3)=v*v*0.5
    r1(1)=1.; r1(2)=v-c; r1(3)=h-v*c
    r2(1)=1.; r2(2)=v+c; r2(3)=h+v*c
    SELECT CASE(ischema)
    CASE(0)


       ! Roe scheme
       phi2= phip(vp(1),c)* l(1)*r0 + phip(vp(2),c)* l(2)* r1+phip(vp(3),c)*l(3)*r2
       phi1= phim(vp(1),c)* l(1)*r0 + phim(vp(2),c)* l(2)* r1+phim(vp(3),c)*l(3)*r2
       ! schema LxF

    CASE(1)
       !Lxf
       ph(1)= u2(2)-u1(2)
       ph(2)= v2*u2(2)-v1*u1(2)+(p2-p1)
       ph(3)= v2*(u2(3)+p2)-v1*(u1(3)+p1)
       alpha=.5*MAXVAL(ABS(vp))
       phi1=0.5*ph-alpha*(u2-u1)
       phi2=0.5*ph+alpha*(u2-u1)
    CASE default
       ! Roe
    END SELECT
  END SUBROUTINE phi


  FUNCTION ap(Ub,gamma) RESULT(a)
    ! calcule la partie positive de A (matrice)
    IMPLICIT NONE
    REAL , DIMENSION(3),INTENT(in) :: ub
    REAL,INTENT(in) :: gamma
    REAL, DIMENSION(3,3):: a,b,d
    REAL, DIMENSION(3,3):: formes_propres
    REAL, DIMENSION(3,3):: vect_propres
    REAL, DIMENSION(3,1):: ligne
    REAL, DIMENSION(1,3):: col
    REAL, DIMENSION(3) :: valp,x,y
    REAL               :: ro,v,p, h, c2, c
    INTEGER :: i

    ro=ub(1); v=ub(2)/ub(1); p=(gamma-1)*(Ub(3)-.5*ub(2)*ub(2)/ro)

    H=(ub(3)+p)/ro
    c2=(gamma-1)*(H-.5*v*v)
    c=SQRT(c2)
    valp(1)=v; valp(2)=v-c; valp(3)=v+c
    vect_propres  =v_propres(ub,gamma)
    formes_propres=f_propres(ub,gamma)

    a=0; d=0

    DO i=1,3

       col(1,:)  =formes_propres(i,:) ! convention d'écriture dur Fortran
       ligne(:,1)=vect_propres(i,:)!vect_propres(i,:) ! convention d'écriture dur Fortran
       d=MATMUL(ligne,col)
       a= a+phip(valp(i),c)*d

    ENDDO
    !print*,a
    a=TRANSPOSE(a)
  END FUNCTION ap

  FUNCTION am(Ub,gamma) RESULT(a)
    ! calcule la partie negative de A (matrice)
    IMPLICIT NONE
    REAL , DIMENSION(3),INTENT(in) :: ub
    REAL,INTENT(in) :: gamma
    REAL, DIMENSION(3,3):: a,b
    REAL, DIMENSION(3,3):: formes_propres
    REAL, DIMENSION(3,3):: vect_propres,d
    REAL, DIMENSION(3,1):: ligne
    REAL, DIMENSION(1,3):: col
    REAL, DIMENSION(3) :: valp
    REAL               :: ro,v,p, h, c2, c
    INTEGER :: i


    ro=ub(1); v=ub(2)/ub(1); p=(gamma-1)*(Ub(3)-.5*ub(2)*ub(2)/ro)

    H=(ub(3)+p)/ro
    c2=(gamma-1)*(H-.5*v*v)
    c=SQRT(c2)
    valp(1)=v; valp(2)=v-c; valp(3)=v+c

    vect_propres  =v_propres(ub,gamma)
    formes_propres=f_propres(ub,gamma)

    a=0; d=0

    DO i=1,3

       col(1,:)  =formes_propres(i,:) ! convention d'écriture dur Fortran
       ligne(:,1)=vect_propres(i,:)!vect_propres(i,:) ! convention d'écriture dur Fortran
       d=MATMUL(ligne,col)!!$    PRINT*, matmul(a,y)


       a= a+ phim(valp(i),c)*d

    ENDDO
    a=TRANSPOSE(a)


  END FUNCTION am


  FUNCTION v_propres (ub,gamma) RESULT(vect) ! vecteurs propres 
    IMPLICIT NONE
    REAL, DIMENSION(3), INTENT(in) :: ub
    REAL, INTENT(in) :: gamma
    REAL, DIMENSION(3,3):: vect ! 1: vp u, 2 : vp u-c, 3: vp u+c
    REAL :: c,h,q2,ro,u,E,p,c2
    ro=ub(1)
    u=ub(2)/ro
    q2=u*u
    E=ub(3)
    p=(gamma-1)*(E-.5*ro*q2)
    H=(E+p)/ro
    c2=(gamma-1)*(H-.5*q2)
    c=SQRT(c2)

    vect(1,1)=1
    vect(1,2)=u
    vect(1,3)=q2*.5

    vect(2,1)=1
    vect(2,2)=u-c
    vect(2,3)=h-u*c

    vect(3,1)=1
    vect(3,2)=u+c
    vect(3,3)=h+u*c

  END FUNCTION v_propres

  FUNCTION  f_propres (ub,gamma) RESULT(vect) ! formes propres
    IMPLICIT NONE
    REAL, DIMENSION(3), INTENT(in) :: ub
    REAL, INTENT(in) :: gamma
    REAL, DIMENSION(3,3):: vect ! 1: vp u, 2 : vp u-c, 3: vp u+c
    REAL :: c,h,q2,ro,u,E,p,c2, kappa
    kappa=gamma-1
    ro=ub(1)
    u=ub(2)/ro
    q2=u*u
    E=ub(3)
    p=kappa*(E-.5*ro*q2)
    H=(E+p)/ro
    c2=kappa*(H-.5*q2)
    c=SQRT(c2)

    vect(1,1)= kappa*q2/c2*.5
    vect(1,2)=- kappa*u/c2
    vect(1,3)= kappa/c2

    vect(2,1)=-u/c
    vect(2,2)=1./c
    vect(2,3)=0.


    vect(3,:)=vect(2,:)

    vect(2,:)=0.5*(vect(1,:)-vect(2,:))
    vect(3,:)=0.5*(vect(1,:)+vect(3,:))
    vect(1,:)=-vect(1,:)
    vect(1,1)=1.+vect(1,1)

  END FUNCTION f_propres

  REAL FUNCTION vmax(n,gamma,u)
    ! 
    ! compute maximal speed 9for CFL)
    !
    IMPLICIT NONE
    INTEGER, INTENT(in):: n
    REAL, INTENT(in):: gamma
    REAL, DIMENSION(3,0:n+1),INTENT(in) ::u
    REAL:: p, v, ro, c, v0
    INTEGER :: i
    v0=-1
    DO i=0,n+1
       ro=u(1,i)
       v=u(2,i)/ro
       p=(gamma-1)*( u(3,i)-0.5*ro*v*v)
       c=SQRT( gamma* p/ro)
       v0=MAX(v0,ABS(v)+c)
    ENDDO
    vmax=v0
  END FUNCTION vmax

  FUNCTION moyenne(U1,U2,gamma)  RESULT (Ub)
    ! Roe average. Could be arithmetic average
    IMPLICIT NONE
    REAL, DIMENSION(3), INTENT(in) :: U1,U2
    REAL, INTENT(in) :: gamma
    REAL, DIMENSION(3):: Ub
    REAL ,DIMENSION(2):: z1,z2,z3
    REAL :: zm1,zm2,zm3
    REAL :: p
    z1(1)=SQRT(U1(1))
    z2(1)=U1(2)/z1(1)
    p= (gamma-1)*( U1(3)-.5*U1(2)*U1(2)/U1(1))
    z3(1)=( U1(3)+p)/z1(1)


    z1(2)=SQRT(U2(1))
    z2(2)=U2(2)/z1(2)
    p= (gamma-1)*( U2(3)-.5*U2(2)*U2(2)/U2(1))
    z3(2)=( U2(3)+p)/z1(2)

    zm1=0.5*(z1(1)+z1(2))
    zm2=0.5*(z2(1)+z2(2))
    zm3=0.5*(z3(1)+z3(2))

    Ub(1)=zm1*zm1
    Ub(2)=zm1*zm2
    Ub(3)= zm1*zm3/gamma+ .5*(gamma-1)/gamma*zm2*zm2
  END FUNCTION moyenne
  REAL FUNCTION f1(x)
    IMPLICIT NONE
    REAL ::x
    IF (ABS(x).LE.1.e-10) THEN
       f1=0
    ELSE
       f1=-MIN(x,0.)/ABS(x)
    ENDIF
  END FUNCTION f1

  REAL FUNCTION mymax(a,b)
    IMPLICIT NONE
    REAL :: a,b
    REAL:: eps=1.e-10
    mymax=MAX(a,b)
  END FUNCTION mymax

  REAL FUNCTION myabs(x,c)
    IMPLICIT NONE
    REAL :: x,eps,c
    eps=1.e-2*c ! correction entropique
    IF (ABS(x).LE.eps) THEN
       myabs=(x*x+eps*eps)/(2*eps)
    ELSE
       myabs=ABS(x)
    ENDIF
  END FUNCTION myabs

  REAL FUNCTION phip(x,c)
    IMPLICIT NONE
    REAL :: x,eps,c
    eps=1.e-1*c! correction entropique
    IF (ABS(x).LE.eps) THEN
       phip=0.5* (x+eps)*(x+eps)/(2*eps)
    ELSE
       phip=MAX(x,0.)
    ENDIF
  END FUNCTION phip

  REAL FUNCTION phim(x,c)
    IMPLICIT NONE
    REAL :: x,eps,c
    eps=1.e-1*c! correction entropique
    IF (ABS(x).LE.eps) THEN
       phim=-0.5* (x-eps)*(x-eps)/(2*eps)
    ELSE
       phim=MIN(x,0.)
    ENDIF
  END FUNCTION phim


END MODULE Stroumpf
