MODULE utils
  ! R. Abgrall
  ! University of Zurich
  ! remi.abgrall@math.uzh.ch
  PRIVATE
  PUBLIC:: inverse
CONTAINS


  FUNCTION inverse( Mat)  RESULT (Mat1)
    !------------------------------------------
    !  Inversion d'une matrice carree par LU et pivot partiel par ligne
    !  Ref : Numerical reciepes in C
    ! -----------------------------------------
    IMPLICIT NONE
    REAL, DIMENSION(3,3), INTENT(in) :: Mat
    REAL, DIMENSION(3,3) :: Mat1, mat2
    REAL, DIMENSION(3)   :: V, COL
    REAL                        :: d
    INTEGER    :: i,j, Nsize
    INTEGER, DIMENSION(4) :: INDX

    IF( SIZE(Mat,DIM=1) .NE. SIZE(Mat,DIM=2) ) THEN
       STOP
    END IF

    Nsize = SIZE(Mat,DIM=1)
    Mat2  = Mat

    CALL LUDCMP(MAT2, NSIZE, INDX, D)
    DO j =  1, Nsize
       col=0.
       col(j)=1.
       CALL luksb(MAT2, Nsize, indx, col)
       Mat1(:,j)=col(:)
    ENDDO

  CONTAINS

    SUBROUTINE LUDCMP(MAT2, NSIZE, INDX, D)
      INTEGER i,j,Nsize,k,imax
      REAL, DIMENSION(Nsize, Nsize) :: Mat2
      REAL, DIMENSION(Nsize) :: vv
      INTEGER, DIMENSION(Nsize) :: Indx
      REAL :: d, Big, dum,sum, temp

      d=1.0
      DO i=1,Nsize
         big=0
         DO j=1,nsize
            temp=ABS(mat2(i,j))
            IF (temp > big) big=temp
         ENDDO
         IF (big == 0.0) THEN
            WRITE(*,*) 'Matrice singuliere LUDCMP'
            STOP
         ENDIF
         vv(i)=1./big
      ENDDO

      DO j=1, Nsize
         DO i=1, j-1
            sum=mat2(i,j)
            DO k=1,i-1
               sum=sum-mat2(i,k)*mat2(k,j)
            ENDDO
            mat2(i,j)=sum
         ENDDO

         big=0
         DO i=j, nsize
            sum=mat2(i,j)
            DO k=1,j-1
               sum=sum-mat2(i,k)*mat2(k,j)
            ENDDO
            mat2(i,j)=sum
            dum=vv(i)*ABS(sum)
            IF (dum .GE. big) THEN
               big=dum
               imax=i
            ENDIF
         ENDDO


         IF (j .NE. imax) THEN
            DO k=1,nsize
               dum=mat2(imax,k)
               mat2(imax,k)=mat2(j,k)
               mat2(j,k)=dum
            ENDDO
            d=-d
            vv(imax)=vv(j)
         ENDIF

         indx(j)=imax
         IF (mat2(j,j)==0.) mat2(j,j)=1.e-20
         IF (j .NE. nsize) THEN
            dum=1./mat2(j,j)
            DO i=j+1,nsize
               mat2(i,j)=mat2(i,j)*dum
            ENDDO
         ENDIF
      END DO
    END SUBROUTINE LUDCMP


    SUBROUTINE LUKSb(Mat2, Nsize, Indx, col)
      INTEGER nsize
      REAL,  DIMENSION(nsize,nsize) :: mat2
      INTEGER,  DIMENSION(nsize) :: indx
      REAL, DIMENSION(nsize) :: col
      INTEGER ::  i,ii=1,ip,j
      REAL  :: sum

      DO i=1,nsize
         ip=indx(i)
         sum=col(ip)
         col(ip)=col(i)
         IF (ii .GT. 0) THEN
            DO j=ii,i-1
               sum=sum-mat2(i,j)*col(j)
            ENDDO
         ELSE
            IF (sum.EQ.0) ii=i
         ENDIF

         col(i)=sum
      ENDDO

      DO i=nsize,1,-1
         sum=col(i)
         DO j=i+1,nsize
            sum=sum-mat2(i,j)*col(j)
         ENDDO
         col(i)=sum/mat2(i,i)
      ENDDO
    END SUBROUTINE  LUKSb
  END FUNCTION INVERSE



END MODULE utils



